#!/usr/bin/python

import random
import string
import os
import subprocess
import datetime
import shutil
import time
import hashlib
import sys
import ssl
from ftplib import FTP
from ftplib import FTP_TLS


'''
1. 打包备份
2. 上传备份
'''

class BackupLVMUtils(object):
    def __init__(self):
        random_str = ''.join(random.sample(string.ascii_lowercase, 4))
        self.suffix_snapname = '.tmpsnap' + random_str
        # self.prefix_block_device = "/dev"
        self.prefix_backup_datadir = "/cloudclusters/backup"
        self.temp_mount_datadir = '/mnt'
        self.temp_mount_datadir_suffix = '.tmp' + random_str


    # 判断 LVM 逻辑卷是否存在
    def exists_lvm(self, lvm_fullname=""):
        if os.path.exists(lvm_fullname.rstrip('/')):
            return True
        else:
            return False


    # 创建存放备份数据的目录
    def create_backupdir(self, stackname=""):
        backup_dir = self.prefix_backup_datadir.rstrip('/') + '/' + stackname
        try:
            if not os.path.exists(backup_dir):
                os.makedirs(backup_dir)
            return True
        except Exception as e:
            print(e)
            return False


    # 创建临时的数据挂载目录 （用来挂载，备份打包）
    def create_temp_mount_datadir(self, stackname=""):
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + stackname + self.temp_mount_datadir_suffix
        try:
            if not os.path.exists(temp_mount_datadir):
                os.makedirs(temp_mount_datadir)
            return True
        except Exception as e:
            print(e)
            return False

    # 创建 LVM 快照
    def create_lvm_snapshot(self, stackname="", lvm_fullname="", lvm_size=""):
        snapshot_name = stackname + self.suffix_snapname
        # cmd = 'lvcreate -L {lv_snap_size} --snapshot --name {lv_snap_name} {lv_origin_name} ; echo $?'.format(
        #     lv_snap_size=lvm_size,
        #     lv_snap_name=snapshot_name,
        #     lv_origin_name=lvm_fullname)
        cmd = 'lvcreate --snapshot --name {lv_snap_name} {lv_origin_name} ; echo $?'.format(
            lv_snap_size=lvm_size,
            lv_snap_name=snapshot_name,
            lv_origin_name=lvm_fullname)
        res = subprocess.getstatusoutput(cmd)
        res_code = res[0]
        res_content = res[1]
        if int(res_code) == 0:
            # 激活
            lv_full_snapname = lvm_fullname.rstrip('/') + self.suffix_snapname
            print(lv_full_snapname)
            cmd2 = 'lvchange -ay -K {lv_full_snapname} ; echo $?'.format(lv_full_snapname=lv_full_snapname)
            res = subprocess.getstatusoutput(cmd2)
            res_code = res[0]
            res_content = res[1]
            if int(res_code) == 0:
                return True
            else:
                return False
        else:
            return False

    # 判断 LVM 逻辑卷快照是否存在
    def exists_lvm_snapshot(self, lvm_fullname=""):
        lvm_snapshot_fullname = lvm_fullname.rstrip('/') + self.suffix_snapname
        if os.path.exists(lvm_snapshot_fullname):
            return True
        else:
            return False


    # 挂载 LVM 逻辑卷到临时的数据目录
    def mount_lvm_to_temp_datadir(self, stackname="", lvm_fullname=""):
        lvm_snapshot_fullname = lvm_fullname.rstrip('/') + self.suffix_snapname
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + stackname + self.temp_mount_datadir_suffix
        cmd = "mount -o discard {block_device} {mount_path}".format(block_device=lvm_snapshot_fullname,
                                                                    mount_path=temp_mount_datadir)
        res = subprocess.getstatusoutput(cmd)
        res_code = res[0]
        res_content = res[1]
        if int(res_code) == 0:
            return True
        else:
            print('cmd execute error: return code: {}, error: {}'.format(res_code, res_content))
            return False


    def pack_filedir_to_backupdir(self, stackname="", timeout=7200, backup_filename=None):
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + stackname + self.temp_mount_datadir_suffix
        backup_datadir = self.prefix_backup_datadir.rstrip('/') + '/' + stackname

        if backup_filename:
            backup_filename = backup_filename
        else:
            backup_file_timeformat = datetime.datetime.now().strftime("%Y%m%d")
            backup_filename = stackname + '.' + backup_file_timeformat + '.tar.gz'

        backup_file_fullpath = backup_datadir.rstrip('/') + '/' + backup_filename

        cmd = 'tar -pczvf {backup_file_fullpath}  -C {data_dir} . '.format(backup_file_fullpath=backup_file_fullpath,
                                                                           data_dir=temp_mount_datadir)
        try:
            # res = subprocess.call(cmd, shell=True, timeout=timeout)
            # if int(res) == 0:
            #     return True
            # else:
            #     return False
            res = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                 universal_newlines=True, timeout=timeout)
            # logger.info('stdout: {}'.format(res.stdout))
            # logger.info('stderr: {}'.format(res.stderr))
            if res.returncode == 0:
                return True
            else:
                print('stdout: {}'.format(res.stdout))
                print('stderr: {}'.format(res.stderr))
                return False
        except Exception as e:
            print(str(e))
            print("export backup file timeout")
            return False


    def exists_backup_file(self, stackname="", backup_filename=""):
        backup_datadir = self.prefix_backup_datadir.rstrip('/') + '/' + stackname

        if backup_filename:
            backup_filename = backup_filename
        else:
            backup_file_timeformat = datetime.datetime.now().strftime("%Y%m%d")
            backup_filename = stackname + '.' + backup_file_timeformat + '.tar.gz'
        backup_file_fullpath = backup_datadir.rstrip('/') + '/' + backup_filename
        if os.path.exists(backup_file_fullpath):
            return True
        else:
            return False


    def umount_temp_datadir(self, stackname=""):
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + stackname + self.temp_mount_datadir_suffix
        cmd = "umount {mount_path}".format(mount_path=temp_mount_datadir)

        res = subprocess.getstatusoutput(cmd)
        res_code = res[0]
        res_content = res[1]
        if int(res_code) == 0:
            return True
        else:
            print(res_content)
            print(res_content)
            return False


    def remove_lvm_snapshot(self, lvm_fullname=""):
        lvm_snaphost_fullname = lvm_fullname.rstrip('/') + self.suffix_snapname
        cmd = "lvremove {lv_name} -y".format(lv_name=lvm_snaphost_fullname)
        res = subprocess.getstatusoutput(cmd)
        res_code = res[0]
        res_content = res[1]
        if int(res_code) == 0:
            return True
        else:
            print(res_content)
            print(res_content)
            return False


    def path_is_mounted(self, data_path):
        return os.path.ismount(data_path)


    def get_lvm_info(self, stackname=""):
        cmd = 'lvs  -o lv_size,lv_full_name 2>/dev/null |grep -w {stackname}|head -n 1'.format(stackname=stackname)
        res = subprocess.getstatusoutput(cmd)
        res_code = res[0]
        if res[1] == '':
            print("LVM [{lvm_name}] 不存在".format(lvm_name=stackname))
            print("LVM [{lvm_name}] 不存在".format(lvm_name=stackname))
        else:
            res_content = res[1].strip().split()
            if int(res_code) == 0:
                lvm_name = res_content[1]
                lvm_size = res_content[0]
                kwargs = {
                    "lvm_size": lvm_size,
                    "lvm_name": lvm_name,
                }
                return kwargs


class AtomError(Exception):
    def __init__(self, msg):
        self.message = msg

    def __str__(self):
        return self.message


class BackupLVM(BackupLVMUtils):
    def __init__(self, stackname="", backup_filename=""):
        super().__init__()
        lvm_info = self.get_lvm_info(stackname=stackname)
        lvm_size = lvm_info["lvm_size"]
        lvm_name = lvm_info["lvm_name"]
        lvm_fullname = '/dev/' + lvm_name
        # 执行顺序
        self.change_list = ["first", "second", "third", "fourth", "fifth", "sixth", "seventh", "eighth", "ninth",
                            "tenth", "eleventh"]
        # 保存每一步old、new状态
        self.states = []
        self.ret = {"result": None}

        self.lvm_fullname = lvm_fullname
        self.stackname = stackname
        self.lvm_size = lvm_size
        self.backup_filename = backup_filename

        self.backup_file_fullpath = os.path.join(self.prefix_backup_datadir, self.stackname, self.backup_filename)

    def start_backup_lvm(self):
        try:
            backup_res = self.atombackup()
            return backup_res
        except Exception as e:
            print(e)
            return False

    # 设置备份超时秒数
    def atombackup(self):
        # 维护一个执行轨迹的 exec_tail 列表
        # 执行时，执行完一个往 exec_tail 中加一个
        # 回滚时，回滚一个从 exec_tail 列表中 pop 掉一个
        start_time = time.time()

        exec_tail = []
        exec_ret, undo_ret = "success", ""
        exec_step = 0
        error_msg = ''
        for n, exec in enumerate(self.change_list):
            try:
                # 模拟原子中某个不止执行失败
                # if n == 3:
                #     raise AtomError("Just a test")
                target = getattr(self, exec + "_step", None)
                target()
                # print(exec + "_step")
            except Exception as e:
                # 记录日志
                exec_ret = "fail"
                error_msg = str(e)
                # print("exec_ret: {}".format(exec_ret))
                # 跳出for循环
                break
            finally:
                exec_tail.append(exec)
            # 如果要对最后一次执行失败的步骤进行状态校验就用这个替换else分支
            # finally:
            # 上面的break触发后，finally下的命令还是会执行的
            # exec_tail.append(exec)

        print(exec_tail)

        if exec_ret == "fail":
            undo_ret = "successful"
            while (len(exec_tail) > 0):
                last_step = exec_tail.pop()
                # print(last_step,'---')
                try:
                    target = getattr(self, last_step + "_undo", None)
                    target()
                except Exception as e:
                    # 记录日志并退出for循环
                    undo_ret = "fail"
                    error_msg += ", rollback failed: " + str(e)
                    break
        backup_file = {}
        if exec_ret == 'success':
            backup_file = {'file_path': self.backup_file_fullpath}

        self.ret = {"result": exec_ret, "undo_ret": undo_ret, "backup_file": backup_file, "error_msg": error_msg}
        end_time = time.time()
        print("耗时: {:.2f}秒".format(end_time - start_time))
        return self.ret

    # 1、 先判断备份文件是否存在，存在则当天不备份
    def first_step(self):
        backup_datadir = os.path.join(self.prefix_backup_datadir, self.stackname)

        if self.backup_filename:
            backup_filename = self.backup_filename
        else:
            # 不传文件名则用平台默认名
            backup_file_timeformat = datetime.datetime.now().strftime("%Y%m%d")
            backup_filename = self.stackname + '.' + backup_file_timeformat + '.tar.gz'

        backup_file_fullpath = backup_datadir.rstrip('/') + '/' + backup_filename

        backup_file_exists = self.exists_backup_file(stackname=self.stackname, backup_filename=backup_filename)

        if not backup_file_exists:
            log_info1 = "1. 备份文件 [{backup_file_fullpath}] 不存在，可以进行备份".format(backup_file_fullpath=backup_file_fullpath)
            print(log_info1)
        else:
            log_info1 = "1. 备份文件 [{backup_file_fullpath}] 存在，不能重复备份".format(backup_file_fullpath=backup_file_fullpath)
            print(log_info1)
            raise AtomError(f"1. Backup file [{backup_file_fullpath}] exists, can't backup again")

    # 1、 回退，如果备份文件存在，则回退（不做任何操作）
    def first_undo(self):
        log_info1 = "1. [回退成功，不做任何操作]"
        print(log_info1)

        backup_datadir = self.prefix_backup_datadir.rstrip('/') + '/' + self.stackname

        if self.backup_filename:
            backup_filename = self.backup_filename
        else:
            # 不传文件名则用平台默认名
            backup_file_timeformat = datetime.datetime.now().strftime("%Y%m%d")
            backup_filename = self.stackname + '.' + backup_file_timeformat + '.tar.gz'

        backup_file_fullpath = backup_datadir.rstrip('/') + '/' + backup_filename

        # 删除备份文件
        if os.path.exists(backup_file_fullpath):
            try:
                os.remove(backup_file_fullpath)
                log_info1 = "1. 删除打包压缩后的数据文件 [{backup_file_fullpath}] 成功 [回退成功]".format(
                    backup_file_fullpath=backup_file_fullpath)
                print(log_info1)
            except:
                log_info1 = "1. 删除打包压缩后的数据文件 [{backup_file_fullpath}] 失败 [回退失败]".format(
                    backup_file_fullpath=backup_file_fullpath)
                print(log_info1)
                raise AtomError(f"1. Delete backup file [{backup_file_fullpath}] failed [undo failed]")

    # 2、先判断 LVM 逻辑卷是否存在该节点，不存在则没法进行备份
    def second_step(self):
        is_lvm_existed = self.exists_lvm(lvm_fullname=self.lvm_fullname)
        if is_lvm_existed:
            log_info2 = "2. LVM [{lvm_fullname}] 存在于该主机，开始备份".format(lvm_fullname=self.lvm_fullname)
            print(log_info2)
        else:
            log_info2 = "2. LVM [{lvm_fullname}] 不存在于该主机，无法备份".format(lvm_fullname=self.lvm_fullname)
            print(log_info2)
            raise AtomError(f"2. LVM [{self.lvm_fullname}] not existed, can't backup")

    # 2、 回退，如果 LVM 不存在改主机，则回退 （不做任何操作）
    def second_undo(self):
        log_info2 = "2. [回退成功，不做任何操作]"
        print(log_info2)

    # 3、 LVM 逻辑卷快照要不存在
    def third_step(self):
        lvm_snapshot_fullname = self.lvm_fullname.rstrip('/') + self.suffix_snapname

        is_lvm_snap_existed = self.exists_lvm_snapshot(lvm_fullname=self.lvm_fullname)
        if not is_lvm_snap_existed:
            log_info3 = "3. LVM 快照 [{lvm_snap_name}] 不存在，准备创建".format(lvm_snap_name=lvm_snapshot_fullname)
            print(log_info3)
        else:
            log_info3 = "3. LVM 快照 [{lvm_snap_name}] 已经存在，需要手动清理".format(lvm_snap_name=lvm_snapshot_fullname)
            print(log_info3)
            raise AtomError(f"3. LVM snapshot [{lvm_snapshot_fullname}] existed, please clean it manually")

    # 3、回滚，逻辑卷快照存在，第三步不做操作，后面有删除的流程
    def third_undo(self):
        lvm_snapshot_fullname = self.lvm_fullname.rstrip('/') + self.suffix_snapname

        log_info3 = "3. 使逻辑卷快照不存在 [回退成功]".format(lvm_snap_name=lvm_snapshot_fullname)
        print(log_info3)

        is_lvm_snap_existed = self.exists_lvm_snapshot(lvm_fullname=self.lvm_fullname)
        if is_lvm_snap_existed:
            # 判断临时数据目录是否已经挂载
            temp_mount_datadir = self.temp_mount_datadir.rstrip(
                '/') + '/' + self.stackname + self.temp_mount_datadir_suffix

            if self.path_is_mounted(data_path=temp_mount_datadir):
                umounted_temp_datadir = self.umount_temp_datadir(stackname=self.stackname)
                if umounted_temp_datadir:
                    log_info3 = "3. 卸载临时数据目录挂载点 [{temp_mount_datadir}] 成功 [回退成功]".format(
                        temp_mount_datadir=temp_mount_datadir)
                    print(log_info3)
                else:
                    log_info3 = "3. 卸载临时数据目录挂载点 [{temp_mount_datadir}] 失败 [回退失败]".format(
                        temp_mount_datadir=temp_mount_datadir)
                    print(log_info3)
                    raise AtomError(f"3. Umount temp mount datadir [{temp_mount_datadir}] failed [undo failed]")

            lvm_snap_removed = self.remove_lvm_snapshot(lvm_fullname=self.lvm_fullname)

            if lvm_snap_removed:
                log_info3 = "3. 删除 LVM 快照 [{lvm_snap_name}] 成功  [回退成功]".format(lvm_snap_name=lvm_snapshot_fullname)
                print(log_info3)
            else:
                log_info3 = "3. 删除 LVM 快照 [{lvm_snap_name}] 失败  [回退失败]".format(lvm_snap_name=lvm_snapshot_fullname)
                print(log_info3)
                raise AtomError(f"3. Remove LVM snapshot [{lvm_snapshot_fullname}] failed [undo failed]")

    # 4、 创建 LVM 快照
    def fourth_step(self):
        lvm_snapshot_fullname = self.lvm_fullname.rstrip('/') + self.suffix_snapname
        lvm_created = self.create_lvm_snapshot(stackname=self.stackname, lvm_fullname=self.lvm_fullname,
                                               lvm_size=self.lvm_size)
        if lvm_created:
            log_info4 = "4. 创建 LVM 快照 [{lvm_snap_name}] 成功".format(lvm_snap_name=lvm_snapshot_fullname)
            print(log_info4)
        else:
            log_info4 = "4. 创建 LVM 快照 [{lvm_snap_name}] 失败".format(lvm_snap_name=lvm_snapshot_fullname)
            print(log_info4)
            raise AtomError(f"4. Create LVM snapshot [{lvm_snapshot_fullname}] failed")

    # 4、 回退， 创建 LVM 快照， 回退不做任何操作
    def fourth_undo(self):
        lvm_snapshot_fullname = self.lvm_fullname.rstrip('/') + self.suffix_snapname

        log_info4 = "4. 使逻辑卷快照 [{lvm_snap_name}] 不存在 [回退成功]".format(lvm_snap_name=lvm_snapshot_fullname)
        print(log_info4)

        is_lvm_snap_existed = self.exists_lvm_snapshot(lvm_fullname=self.lvm_fullname)
        if is_lvm_snap_existed:
            lvm_snap_removed = self.remove_lvm_snapshot(lvm_fullname=self.lvm_fullname)

            if lvm_snap_removed:
                log_info4 = "4. 删除 LVM 快照 [{lvm_snap_name}] 成功  [回退成功]".format(lvm_snap_name=lvm_snapshot_fullname)
                print(log_info4)
            else:
                log_info4 = "4. 删除 LVM 快照 [{lvm_snap_name}] 失败  [回退失败]".format(lvm_snap_name=lvm_snapshot_fullname)
                print(log_info4)
                raise AtomError(f"4. Remove LVM snapshot [{lvm_snapshot_fullname}] failed [undo failed]")

    # 5、创建存放 临时快照数据 的目录
    def fifth_step(self):
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + self.stackname + self.temp_mount_datadir_suffix
        temp_mount_datadir_created = self.create_temp_mount_datadir(stackname=self.stackname)
        if temp_mount_datadir_created:
            log_info5 = "5. 创建临时快照数据挂载目录 [{mount_datadir}] 成功".format(mount_datadir=temp_mount_datadir)
            print(log_info5)
        else:
            log_info5 = "5. 创建临时快照数据挂载目录 [{mount_datadir}] 失败".format(mount_datadir=temp_mount_datadir)
            print(log_info5)
            raise AtomError(f"5. Create temp mount datadir [{temp_mount_datadir}] failed")

    # 5、回退，存放 临时快照数据 的目录，不做任何操作
    def fifth_undo(self):
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + self.stackname + self.temp_mount_datadir_suffix
        if os.path.exists(temp_mount_datadir):
            shutil.rmtree(temp_mount_datadir)
        log_info5 = "5. [回退成功，删除临时快照数据目录]"
        print(log_info5)

    # 6、创建存放备份数据的目录
    def sixth_step(self):
        backup_datadir = self.prefix_backup_datadir.rstrip('/') + '/' + self.stackname
        backupdir_created = self.create_backupdir(stackname=self.stackname)
        if backupdir_created:
            log_info6 = "6. 创建备份数据目录 [{backup_datadir}] 成功".format(backup_datadir=backup_datadir)
            print(log_info6)
        else:
            log_info6 = "6. 创建备份数据目录 [{backup_datadir}] 失败".format(backup_datadir=backup_datadir)
            print(log_info6)
            raise AtomError(f"6. Create backup datadir [{backup_datadir}] failed")

    # 6、回退， 创建存放备份数据的目录失败则不进行任何操作
    def sixth_undo(self):
        log_info6 = "6. [回退成功，不做任何操作]"
        print(log_info6)

    # 7、 挂载 LVM 快照逻辑卷到 临时的数据目录
    def seventh_step(self):
        lvm_snapshot_fullname = self.lvm_fullname.rstrip('/') + self.suffix_snapname
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + self.stackname + self.temp_mount_datadir_suffix
        temp_datadir_is_mounted = self.mount_lvm_to_temp_datadir(stackname=self.stackname,
                                                                 lvm_fullname=self.lvm_fullname)
        if temp_datadir_is_mounted:
            log_info7 = "7. 挂载 LVM 快照逻辑卷 [{lvm_snapshot_fullname}] 到临时数据目录 [{temp_mount_datadir}] 成功".format(
                lvm_snapshot_fullname=lvm_snapshot_fullname, temp_mount_datadir=temp_mount_datadir)
            print(log_info7)
        else:
            log_info7 = "7. 挂载 LVM 快照逻辑卷 [{lvm_snapshot_fullname}] 到临时数据目录 [{temp_mount_datadir}] 失败".format(
                lvm_snapshot_fullname=lvm_snapshot_fullname, temp_mount_datadir=temp_mount_datadir)
            print(log_info7)
            raise AtomError(
                f"7. Mount LVM snapshot [{lvm_snapshot_fullname}] to temp mount datadir [{temp_mount_datadir}] failed")

    # 7、 回退，卸载 LVM 快照逻辑卷到
    def seventh_undo(self):
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + self.stackname + self.temp_mount_datadir_suffix
        log_info7 = "7. 使临时数据目录挂载 [{temp_mount_datadir}] 不存在 [回退成功]".format(temp_mount_datadir=temp_mount_datadir)
        print(log_info7)

        if self.path_is_mounted(data_path=temp_mount_datadir):
            for i in range(10):
                fuser_mount = subprocess.getstatusoutput("fuser " + temp_mount_datadir)
                if fuser_mount[0] == 0:
                    time.sleep(10)
                else:
                    break

            umounted_temp_datadir = self.umount_temp_datadir(stackname=self.stackname)
            if umounted_temp_datadir:
                log_info7 = "7. 卸载临时数据目录挂载点 [{temp_mount_datadir}] 成功 [回退成功]".format(
                    temp_mount_datadir=temp_mount_datadir)
                print(log_info7)
            else:
                log_info7 = "7. 卸载临时数据目录挂载点 [{temp_mount_datadir}] 失败 [回退失败]".format(
                    temp_mount_datadir=temp_mount_datadir)
                print(log_info7)
                raise AtomError(f"7. Umount temp mount datadir [{temp_mount_datadir}] failed [undo failed]")

    # 8、打包压缩文件到 备份数据目录
    def eighth_step(self):
        backup_datadir = self.prefix_backup_datadir.rstrip('/') + '/' + self.stackname
        backup_file_fullpath = backup_datadir.rstrip('/') + '/' + self.backup_filename
        pack_file_success = self.pack_filedir_to_backupdir(stackname=self.stackname, timeout=10800,
                                                           backup_filename=self.backup_filename)
        if pack_file_success:
            log_info8 = "8. 打包压缩数据文件 [{backup_file_fullpath}] 到备份目录 成功".format(
                backup_file_fullpath=backup_file_fullpath)
            print(log_info8)
        else:
            log_info8 = "8. 打包压缩数据文件 [{backup_file_fullpath}] 到备份目录 失败".format(
                backup_file_fullpath=backup_file_fullpath)
            print(log_info8)
            raise AtomError(f"8. Pack filedir to backupdir [{backup_file_fullpath}] failed")

    # 8、删除打包后的备份文件
    def eighth_undo(self):
        backup_datadir = self.prefix_backup_datadir.rstrip('/') + '/' + self.stackname

        if self.backup_filename:
            backup_filename = self.backup_filename
        else:
            # 不传文件名则用平台默认名
            backup_file_timeformat = datetime.datetime.now().strftime("%Y%m%d")
            backup_filename = self.stackname + '.' + backup_file_timeformat + '.tar.gz'

        backup_file_fullpath = backup_datadir.rstrip('/') + '/' + backup_filename
        log_info8 = "8. 使打包压缩后的数据文件 [{backup_file_fullpath}] 不存在 [回退成功]".format(
            backup_file_fullpath=backup_file_fullpath)
        print(log_info8)

        if os.path.exists(backup_file_fullpath):
            try:
                os.remove(backup_file_fullpath)
                log_info8 = "8. 删除打包压缩后的数据文件 [{backup_file_fullpath}] 成功 [回退成功]".format(
                    backup_file_fullpath=backup_file_fullpath)
                print(log_info8)
            except:
                log_info8 = "8. 删除打包压缩后的数据文件 [{backup_file_fullpath}] 失败 [回退失败]".format(
                    backup_file_fullpath=backup_file_fullpath)
                print(log_info8)
                raise AtomError(f"8. Delete backup file [{backup_file_fullpath}] failed [undo failed]")

    # 9、卸载临时数据目录挂载点
    def ninth_step(self):
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + self.stackname + self.temp_mount_datadir_suffix
        umounted_temp_datadir = self.umount_temp_datadir(stackname=self.stackname)
        if umounted_temp_datadir:
            log_info9 = "9. 卸载临时数据目录挂载点 [{temp_mount_datadir}] 成功".format(temp_mount_datadir=temp_mount_datadir)
            print(log_info9)
        else:
            log_info9 = "9. 卸载临时数据目录挂载点 [{temp_mount_datadir}] 失败".format(temp_mount_datadir=temp_mount_datadir)
            print(log_info9)
            raise AtomError(f"9. Umount temp mount datadir [{temp_mount_datadir}] failed")

    # 回退，卸载临时数据目录 挂载点
    def ninth_undo(self):
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + self.stackname + self.temp_mount_datadir_suffix
        log_info9 = "9. 使临时数据目录挂载 [{temp_mount_datadir}] 不存在 [回退成功]".format(temp_mount_datadir=temp_mount_datadir)
        print(log_info9)

        if self.path_is_mounted(data_path=self.temp_mount_datadir):
            umounted_temp_datadir = self.umount_temp_datadir(stackname=self.stackname)
            if umounted_temp_datadir:
                log_info9 = "9. 卸载临时数据目录挂载点 [{temp_mount_datadir}] 成功 [回退成功]".format(
                    temp_mount_datadir=temp_mount_datadir)
                print(log_info9)
            else:
                log_info9 = "9. 卸载临时数据目录挂载点 [{temp_mount_datadir}] 失败 [回退失败]".format(
                    temp_mount_datadir=temp_mount_datadir)
                print(log_info9)
                raise AtomError(f"9. Umount temp mount datadir [{temp_mount_datadir}] failed [undo failed]")

    # 10、移除 LVM 逻辑卷快照
    def tenth_step(self):
        lvm_snaphost_fullname = self.lvm_fullname.rstrip('/') + self.suffix_snapname
        lvm_snap_removed = self.remove_lvm_snapshot(lvm_fullname=self.lvm_fullname)
        if lvm_snap_removed:
            log_info10 = "10. 删除 LVM 逻辑卷快照 [{lvm_snaphost_fullname}] 成功，备份完成 END".format(
                lvm_snaphost_fullname=lvm_snaphost_fullname)
            print(log_info10)
        else:
            log_info10 = "10. 删除 LVM 逻辑卷快照 [{lvm_snaphost_fullname}] 失败，备份失败 END".format(
                lvm_snaphost_fullname=lvm_snaphost_fullname)
            print(log_info10)
            raise AtomError(f"10. Remove LVM snapshot [{lvm_snaphost_fullname}] failed")

    # 10、回退，删除 LVM 快照
    def tenth_undo(self):
        lvm_snapshot_fullname = self.lvm_fullname.rstrip('/') + self.suffix_snapname

        log_info10 = "10. 使逻辑卷快照 [{lvm_snap_name}] 不存在 [回退成功]".format(lvm_snap_name=lvm_snapshot_fullname)
        print(log_info10)

        is_lvm_snap_existed = self.exists_lvm_snapshot(lvm_fullname=self.lvm_fullname)
        if is_lvm_snap_existed:
            lvm_snap_removed = self.remove_lvm_snapshot(lvm_fullname=self.lvm_fullname)

            if lvm_snap_removed:
                log_info10 = "10. 删除 LVM 快照 [{lvm_snap_name}] 成功  [回退成功]".format(lvm_snap_name=lvm_snapshot_fullname)
                print(log_info10)
            else:
                log_info10 = "10. 删除 LVM 快照 [{lvm_snap_name}] 失败  [回退失败]".format(lvm_snap_name=lvm_snapshot_fullname)
                print(log_info10)
                raise AtomError(f"10. Remove LVM snapshot [{lvm_snapshot_fullname}] failed [undo failed]")

    # 11、移除 临时数据挂载目录
    def eleventh_step(self):
        temp_mount_datadir = self.temp_mount_datadir.rstrip('/') + '/' + self.stackname + self.temp_mount_datadir_suffix
        if os.path.exists(temp_mount_datadir):
            shutil.rmtree(temp_mount_datadir)

        log_info11 = "11. 删除逻辑卷快照挂载目录 [{temp_mount_datadir}] 成功，备份完成 END".format(
            temp_mount_datadir=temp_mount_datadir)
        print(log_info11)


def get_absolute_path(path):
    abs_path = os.path.abspath(path)
    if not os.path.exists(abs_path):
        raise FileNotFoundError(f"路径不存在: {abs_path}")
    return abs_path


def generate_remote_path(base_path, file_name):
    if not base_path.endswith('/'):
        base_path += '/'
    return base_path + file_name


class MyFTP(FTP):
    def __init__(self):
        print('make a object')

    def ConnectFTP(self, remote_ip="", remote_port=21, login_name="", login_password=""):
        ftp = FTP_TLS()
        ftp.ssl_version = ssl.PROTOCOL_TLSv1_2

        try:
            ftp.connect(remote_ip, remote_port, 600)
            print('Connect Success')
        except Exception as e:
            print(sys.stderr, "Connect Failed - %s" % e)
            return (0, 'Connect Failed')
        else:
            try:
                ftp.login(login_name, login_password)
                print('Login Success')
            except Exception as e:
                print(sys.stderr, "Login Failed - %s" % e)
                time.sleep(10)
                return (0, 'Login Failed')
            else:
                return (1, ftp)

    def upload(self, remoteHost, remotePort, login_name, login_password, remoteBasePath, localPath):
        localPath = get_absolute_path(localPath)
        localFileName = os.path.basename(localPath)

        remotePath = generate_remote_path(remoteBasePath, localFileName)
        remoteDir, remoteFileName = self.splitpath(remotePath)

        # Connect to FTP
        res = self.ConnectFTP(remote_ip=remoteHost, remote_port=remotePort, login_name=login_name,
                              login_password=login_password)

        if res[0] != 1:
            print(res[1])
            sys.exit()
        ftp = res[1]

        # Ensure the remote directory exists
        try:
            ftp.cwd(remoteDir)
        except Exception:
            # Dynamically create directories
            parts = remoteDir.strip('/').split('/')
            current_path = ''
            for part in parts:
                current_path = f"{current_path}/{part}" if current_path else part
                try:
                    ftp.cwd(current_path)
                except Exception:
                    ftp.mkd(current_path)
                    ftp.cwd(current_path)

        # Check if the remote file exists
        try:
            rsize = ftp.size(remoteFileName)
            print(f"File {remoteFileName} already exists on the server.")
            delete = input("Do you want to delete it and re-upload? (y/n): ").strip().lower()
            if delete != 'y':
                print("Aborting upload. Please delete the file manually if necessary.")
                ftp.quit()
                return
            # If user opts to delete, remove the file on the server
            ftp.delete(remoteFileName)
            print(f"Deleted {remoteFileName} from the server.")
            rsize = 0  # Start upload from the beginning
        except Exception:
            rsize = 0  # File doesn't exist, proceed with upload

        lsize = os.stat(localPath).st_size  # Get local file size
        print(f"Local file size: {lsize} bytes, Remote file size: {rsize} bytes")

        # Upload the file from the current position
        with open(localPath, 'rb') as localfile:
            localfile.seek(rsize)  # Resume from the last position

            uploaded = [rsize]  # Track the uploaded bytes

            def progress_callback(data):
                uploaded[0] += len(data)  # Update uploaded bytes count
                percent = (uploaded[0] / lsize) * 100
                progress_bar = ('#' * int(percent // 2)).ljust(50)  # Every 2% adds a '#'
                print(f"\r[{progress_bar}] {percent:.2f}% ({uploaded[0]}/{lsize} bytes)", end='')

                # Save the uploaded position to a file
                with open(localFileName + ".upload_position", 'w') as f:
                    f.write(str(uploaded[0]))

            # Upload the file with callback to track progress
            ftp.storbinary(f"STOR {remoteFileName}", localfile, callback=progress_callback)

        print(f"\nUploaded {localPath} to {remotePath}")
        ftp.quit()
        os.remove(localFileName + ".upload_position")  # Delete the position file after completion

    def splitpath(self, remotepath):
        position = remotepath.rfind('/')
        return (remotepath[:position + 1], remotepath[position + 1:])

    def md5_calc(self, file):
        md5_value = hashlib.md5()
        with open(file, 'rb') as file_b:
            while True:
                data_flow = file_b.read(8096)
                if not data_flow:
                    break
                md5_value.update(data_flow)
        return md5_value.hexdigest()

    def get_filesize(self, filename):
        fsize = os.path.getsize(filename)
        return fsize



def main():
    stackname = input("Enter the namespace app name to clone: ").strip()
    login_name = input("Enter FTP username: ").strip()
    login_password = input("Enter FTP password: ").strip()
    backup_filename = f"{stackname}-clone.tar.gz"

    backobj = BackupLVM(stackname=stackname, backup_filename=backup_filename)
    backupfilepath = backobj.backup_file_fullpath
    print(backupfilepath)

    if os.path.exists(backupfilepath):
        pass
    else:
        res = backobj.start_backup_lvm()
        print(res)
        time.sleep(10)

    ftpobj = MyFTP()
    ftpobj.upload(
        remoteHost='165.73.249.37',  # Replace with your FTP server IP
        remotePort=10021,  # Replace with your FTP server port
        login_name=login_name,
        login_password=login_password,
        remoteBasePath="/ftp-download/vps-clone/",  # Replace with your remote base path
        localPath=backupfilepath
    )


if __name__ == "__main__":
    try:
        main()
        input("Upload complete. Press Enter to exit...")
    except Exception as e:
        print(f"An error occurred: {e}")
        input("Press Enter to exit...")
